import { Ok, Err, expect_never } from 'safe-types';

var ignoreDupeMap = {
    age: true,
    authorization: true,
    "content-length": true,
    "content-type": true,
    etag: true,
    expires: true,
    from: true,
    host: true,
    "if-modified-since": true,
    "if-unmodified-since": true,
    "last-modified": true,
    location: true,
    "max-forwards": true,
    "proxy-authorization": true,
    referer: true,
    "retry-after": true,
    "user-agent": true,
};
/**
 * Parse headers into the target object.
 *
 * ```
 * const resHeaders =
 * `Date: Wed, 27 Aug 2014 08:58:49 GMT
 * Content-Type: application/json
 * Connection: keep-alive
 * Transfer-Encoding: chunked`
 *
 * parseHeaders(resHeaders)
 * // Output:
 * { date: 'Wed, 27 Aug 2014 08:58:49 GMT',
 *  'content-type': 'application/json',
 *  connection: 'keep-alive',
 *  'transfer-encoding': 'chunked' }
 * ```
 */
function parse(headers, target_mut) {
    if (!headers) {
        return target_mut;
    }
    var line;
    var key;
    var val;
    var i;
    var lead_sp = /^\s*/;
    var trail_sp = /\s*$/;
    for (var _i = 0, _a = headers.split("\n"); _i < _a.length; _i++) {
        line = _a[_i];
        i = line.indexOf(":");
        key = line
            .substr(0, i)
            .replace(lead_sp, "")
            .replace(trail_sp, "")
            .toLowerCase();
        val = line
            .substr(i + 1)
            .replace(lead_sp, "")
            .replace(trail_sp, "");
        // Skip empty keys and defined items in the ignore list.
        if (key == "" || (target_mut[key] !== undefined && ignoreDupeMap[key])) {
            continue;
        }
        if (key === "set-cookie") {
            // Normalize to array
            if (!Array.isArray(target_mut[key])) {
                target_mut[key] = [];
            }
            // TS needs a var alloc to infer type, so force the type
            target_mut[key].push(val);
            continue;
        }
        // These keys can have multiple values
        if (target_mut[key] !== undefined) {
            target_mut[key] += ", " + val;
            continue;
        }
        target_mut[key] = val;
    }
    return target_mut;
}

function encode_query(data) {
    var encoded = [];
    var enc = encodeURIComponent;
    var val;
    var v;
    var k;
    for (var _i = 0, _a = Object.keys(data); _i < _a.length; _i++) {
        k = _a[_i];
        val = data[k];
        if (typeof val == "function" || val === undefined) {
            // Skip QueryObject key iter for non encodable types.
            continue;
        }
        if (Array.isArray(val)) {
            for (var _b = 0, val_1 = val; _b < val_1.length; _b++) {
                v = val_1[_b];
                // Omit writing undefined value.
                if (v === undefined) {
                    continue;
                }
                encoded.push(enc(k) + "[]=" + enc("" + v));
            }
            // Next QueryObject key.
            continue;
        }
        encoded.push(enc(k) + "=" + enc("" + val));
    }
    return encoded.join("&");
}

var ContentType;
(function (ContentType) {
    ContentType["Json"] = "application/json";
    ContentType["Text"] = "text/plain";
    ContentType["Html"] = "text/html";
    ContentType["Form"] = "application/x-www-form-urlencoded";
    ContentType["Multipart"] = "multipart/form-data";
    ContentType["Stream"] = "application/octet-stream";
})(ContentType || (ContentType = {}));

var RequestErrType;
(function (RequestErrType) {
    RequestErrType["HttpStatusErr"] = "HttpStatusErr";
    RequestErrType["XhrErr"] = "XhrErr";
    RequestErrType["Timeout"] = "Timeout";
    RequestErrType["Abort"] = "Abort";
})(RequestErrType || (RequestErrType = {}));
function err_with_matcher(err) {
    return Object.assign(err, {
        match: function (matcher) {
            switch (err.type) {
                case RequestErrType.Abort:
                    return matcher[RequestErrType.Abort](err);
                case RequestErrType.HttpStatusErr:
                    return matcher[RequestErrType.HttpStatusErr](err);
                case RequestErrType.Timeout:
                    return matcher[RequestErrType.Timeout](err);
                case RequestErrType.XhrErr:
                    return matcher[RequestErrType.XhrErr](err);
                default:
                    return expect_never(err, "invalid error type");
            }
        },
    });
}
/**
 * Parse an XHR into a Response object with a payload `T`.
 */
function new_response(xhr) {
    var r = {
        xhr: xhr,
        data: xhr.response,
        status: xhr.status,
        statusText: xhr.statusText,
        get headers() {
            Object.defineProperty(r, "headers", {
                value: parse(xhr.getAllResponseHeaders(), {}),
            });
            return r.headers;
        },
    };
    return r;
}
/**
 * Client is a class that wraps an individual XMLHttpRequest object.
 * A successful request takes the shape `T`, and an unsuccessful
 * request that returns a response (`HttpStatusErr`) takes the
 * shape `E`.
 */
var Client = /** @class */ (function () {
    function Client(method, url) {
        this.method = method;
        this.url = url;
        this.xhr = new XMLHttpRequest();
        this.headers = Object.assign(Object.create(null), Client.DefaultHeaders);
        this.responseType = Client.ResponseType;
        this.timeout = Client.Timeout;
        this.query_added = false;
        this.req_sent = false;
        this.has_aborted = false;
        this.xhr_evt_listeners = [];
        this.upload_evt_listeners = [];
    }
    /**
     * Adds an object of headers via `Object.assign`
     * that will be written to the XHR.
     */
    Client.prototype.addHeaders = function (headers) {
        Object.assign(this.headers, headers);
    };
    /**
     * Sets a new object of headers that will be written to the XHR.
     */
    Client.prototype.setHeaders = function (headers) {
        this.headers = Object.assign(Object.create(null), headers);
    };
    /**
     * addQueryStr adds an already encoded query string to the url.
     *
     * _NOTE: prepends the query string with '?' char._
     */
    Client.prototype.addQueryStr = function (query) {
        if (this.query_added) {
            throw new Error("Cannot add query string twice.");
        }
        this.url += "?";
        this.url += query;
        this.query_added = true;
    };
    /**
     * addQueryObj builds a query string from the object and adds it to the url.
     */
    Client.prototype.addQueryObj = function (query) {
        this.addQueryStr(encode_query(query));
    };
    Client.prototype.setData = function (data) {
        if (data instanceof FormData || data instanceof Document) {
            this.data = data;
            return;
        }
        this.data = JSON.stringify(data);
        this.addHeaders({ "Content-Type": ContentType.Json + ";charset=utf-8" });
    };
    Client.prototype.applyHeaders = function () {
        var header;
        for (header in this.headers) {
            if (!this.headers[header]) {
                continue;
            }
            this.xhr.setRequestHeader(header, this.headers[header]);
        }
    };
    /**
     * Returns a function that will cancel the underlying xhr.
     * If the xhr is underway, this will generate an error with type: `Abort`.
     */
    Client.prototype.getCancelToken = function () {
        var _this = this;
        return function () {
            _this.has_aborted = true;
            _this.xhr.abort();
        };
    };
    /**
     * Register event handlers on the xhr.
     * There's largely no reason to need this since
     * the xhr events are watched internally.
     */
    Client.prototype.addEventListener = function (type, listener, options) {
        this.xhr_evt_listeners.push({
            type: type,
            listener: listener,
            options: options,
        });
    };
    /**
     * Register event handlers on the xhr upload.
     * Useful for progress events on a file upload.
     */
    Client.prototype.addUploadEventListener = function (type, listener, options) {
        this.upload_evt_listeners.push({
            type: type,
            listener: listener,
            options: options,
        });
    };
    Client.prototype.handleReadyStateChange = function (resolve, event) {
        if (this.xhr.readyState !== XMLHttpRequest.DONE || this.has_aborted) {
            return;
        }
        if (this.xhr.status < 200 || this.xhr.status >= 400) {
            resolve(Err(err_with_matcher({
                xhr: this.xhr,
                type: RequestErrType.HttpStatusErr,
                response: new_response(this.xhr),
                event: event,
            })));
        }
        resolve(Ok(new_response(this.xhr)));
    };
    Client.prototype.execute = function (resolve) {
        var _this = this;
        this.xhr.open(this.method, this.url, true);
        this.xhr.responseType = this.responseType;
        this.xhr.timeout = this.timeout;
        this.applyHeaders();
        this.xhr.onreadystatechange = this.handleReadyStateChange.bind(this, resolve);
        this.xhr.addEventListener("timeout", function (event) {
            return resolve(Err(err_with_matcher({
                xhr: _this.xhr,
                type: RequestErrType.Timeout,
                event: event,
            })));
        });
        this.xhr.addEventListener("abort", function (event) {
            return resolve(Err(err_with_matcher({
                xhr: _this.xhr,
                type: RequestErrType.Abort,
                event: event,
            })));
        });
        this.xhr.addEventListener("error", function (event) {
            return resolve(Err(err_with_matcher({
                xhr: _this.xhr,
                type: RequestErrType.XhrErr,
                event: event,
            })));
        });
        for (var _i = 0, _a = this.xhr_evt_listeners; _i < _a.length; _i++) {
            var l = _a[_i];
            this.xhr.addEventListener(l.type, l.listener, l.options);
        }
        for (var _b = 0, _c = this.upload_evt_listeners; _b < _c.length; _b++) {
            var l = _c[_b];
            this.xhr.upload.addEventListener(l.type, l.listener, l.options);
        }
        this.xhr.send(this.data);
    };
    /**
     * send fires off the XHR and returns a promise with the Result.
     * Multiple calls will result in the same return value (wrapped in Promise).
     */
    Client.prototype.send = function () {
        if (this.req_sent) {
            return this.promise;
        }
        this.promise = new Promise(this.execute.bind(this));
        this.req_sent = true;
        return this.promise;
    };
    /**
     * Construct a client manually for a request with a response shaped T,
     * or in the case of a bad request, shaped E.
     *
     * `Client.create` is useful when you need a cancellation token
     * or need to listen to special XHR events (like upload progress).
     */
    Client.create = function (method, url, options) {
        if (options === void 0) { options = {}; }
        var client = new Client(method, url);
        if (options.query) {
            client.addQueryObj(options.query);
        }
        if (options.data) {
            client.setData(options.data);
        }
        if (options.responseType) {
            client.responseType = options.responseType;
        }
        if (options.headers) {
            client.setHeaders(options.headers);
        }
        if (options.timeout) {
            client.timeout = options.timeout;
        }
        return client;
    };
    /**
     * Make a GET request with a response shaped T,
     * or in the case of a bad request, shaped E.
     */
    Client.get = function (url, query) {
        return Client.create("GET", url, { query: query }).send();
    };
    /**
     * Make a POST request with a response shaped T,
     * or in the case of a bad request, shaped E.
     */
    Client.post = function (url, data, options) {
        if (options === void 0) { options = {}; }
        options.data = data;
        return Client.create("POST", url, options).send();
    };
    /**
     * Make a PUT request with a response shaped T,
     * or in the case of a bad request, shaped E.
     */
    Client.put = function (url, data, options) {
        if (options === void 0) { options = {}; }
        options.data = data;
        return Client.create("PUT", url, options).send();
    };
    /**
     * Make a PATCH request with a response shaped T,
     * or in the case of a bad request, shaped E.
     */
    Client.patch = function (url, data, options) {
        if (options === void 0) { options = {}; }
        options.data = data;
        return Client.create("PATCH", url, options).send();
    };
    /**
     * Make a DELETE request with a response shaped T,
     * or in the case of a bad request, shaped E.
     */
    Client.delete = function (url, query) {
        return Client.create("DELETE", url, { query: query }).send();
    };
    /**
     * Sets the type of the response to enable correct parsing of the
     * XHR response.
     */
    Client.setResponseType = function (type) {
        Client.ResponseType = type;
    };
    /**
     * Timeout value in milliseconds.
     */
    Client.Timeout = 0;
    /**
     * The XHR response parsing strategy.
     */
    Client.ResponseType = "json";
    Client.DefaultHeaders = {
        Accept: ContentType.Json,
        "X-Requested-With": "XMLHttpRequest",
    };
    return Client;
}());

var key = "", err = null, compatChecks = Object.create(null);
compatChecks.XMLHttpRequest = "XMLHttpRequest" in window;
compatChecks.Promise = "Promise" in window;
compatChecks.FormData = "FormData" in window;
compatChecks["Array.isArray"] = typeof Array.isArray == "function";
compatChecks["Object.assign"] = typeof Object.assign == "function";
compatChecks["Object.keys"] = typeof Object.keys == "function";
compatChecks["Object.create"] = typeof Object.create == "function";
for (key in compatChecks) {
    if (compatChecks[key]) {
        continue;
    }
    err = new Error("[browser-http-client] A required API is not available in the browser: " + key);
    if (typeof window.onerror == "function") {
        window.onerror(err.message, err.fileName, err.lineNumber, err.columnNumber, err);
    }
    console.error(err);
}

export { Client };
//# sourceMappingURL=http.esm.js.map
