import { Result } from "safe-types";
import * as Headers from "./headers";
import * as QStr from "../encoding/qstr";
import { ContentType } from "./contentType";
export interface RequestOptions {
    query?: QStr.QueryObject;
    data?: RequestData;
    headers?: Headers.HttpHeaders;
    responseType?: XMLHttpRequestResponseType;
    timeout?: number;
}
export declare type RequestData = {
    [key: string]: any;
} | FormData | Document;
export declare type RequestBody = string | FormData | Document;
export interface Response<T = any> {
    data: T;
    status: number;
    statusText: string;
    headers: {
        [field: string]: string | string[];
    };
    xhr: XMLHttpRequest;
}
export declare enum RequestErrType {
    HttpStatusErr = "HttpStatusErr",
    XhrErr = "XhrErr",
    Timeout = "Timeout",
    Abort = "Abort"
}
export declare type Matchable<E> = {
    match: ClientErrMatcher<E>;
};
export declare type ClientErrorStatus<E> = {
    xhr: XMLHttpRequest;
    readonly type: RequestErrType.HttpStatusErr;
    response: Response<E>;
    event: Event;
};
export declare type ClientErrorAbort = {
    xhr: XMLHttpRequest;
    readonly type: RequestErrType.Abort;
    event: Event;
};
export declare type ClientErrorProgress = {
    xhr: XMLHttpRequest;
    readonly type: RequestErrType.Timeout;
    event: ProgressEvent;
};
export declare type ClientErrorError = {
    xhr: XMLHttpRequest;
    readonly type: RequestErrType.XhrErr;
    event: ErrorEvent;
};
export declare type ClientErrorBase<E> = ClientErrorStatus<E> | ClientErrorAbort | ClientErrorProgress | ClientErrorError;
export declare type ClientError<E> = (Matchable<E> & ClientErrorStatus<E>) | (Matchable<E> & ClientErrorAbort) | (Matchable<E> & ClientErrorProgress) | (Matchable<E> & ClientErrorError);
export declare type ErrMatchObj<E, U> = {
    [RequestErrType.HttpStatusErr]: (err: ClientErrorStatus<E>) => U;
    [RequestErrType.XhrErr]: (err: ClientErrorError) => U;
    [RequestErrType.Timeout]: (err: ClientErrorProgress) => U;
    [RequestErrType.Abort]: (err: ClientErrorAbort) => U;
};
export interface ClientErrMatcher<E> {
    <U>(matcher: ErrMatchObj<E, U>): U;
}
export interface PromiseResolver<T> {
    (value: T): T;
}
/**
 * Client is a class that wraps an individual XMLHttpRequest object.
 * A successful request takes the shape `T`, and an unsuccessful
 * request that returns a response (`HttpStatusErr`) takes the
 * shape `E`.
 */
export declare class Client<T, E> {
    private method;
    private url;
    /**
     * Timeout value in milliseconds.
     */
    static Timeout: number;
    /**
     * The XHR response parsing strategy.
     */
    static ResponseType: XMLHttpRequestResponseType;
    static DefaultHeaders: {
        Accept: ContentType;
        "X-Requested-With": string;
    };
    private xhr;
    private headers;
    private data;
    private responseType;
    private timeout;
    private query_added;
    private req_sent;
    private has_aborted;
    private promise;
    private xhr_evt_listeners;
    private upload_evt_listeners;
    constructor(method: string, url: string);
    /**
     * Adds an object of headers via `Object.assign`
     * that will be written to the XHR.
     */
    addHeaders(headers: Headers.HttpHeaders): void;
    /**
     * Sets a new object of headers that will be written to the XHR.
     */
    setHeaders(headers: Headers.HttpHeaders): void;
    /**
     * addQueryStr adds an already encoded query string to the url.
     *
     * _NOTE: prepends the query string with '?' char._
     */
    addQueryStr(query: string): void;
    /**
     * addQueryObj builds a query string from the object and adds it to the url.
     */
    addQueryObj(query: QStr.QueryObject): void;
    private setData;
    private applyHeaders;
    /**
     * Returns a function that will cancel the underlying xhr.
     * If the xhr is underway, this will generate an error with type: `Abort`.
     */
    getCancelToken(): () => void;
    /**
     * Register event handlers on the xhr.
     * There's largely no reason to need this since
     * the xhr events are watched internally.
     */
    addEventListener<K extends keyof XMLHttpRequestEventMap>(type: K, listener: (this: XMLHttpRequest, ev: XMLHttpRequestEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;
    /**
     * Register event handlers on the xhr upload.
     * Useful for progress events on a file upload.
     */
    addUploadEventListener<K extends keyof XMLHttpRequestEventTargetEventMap>(type: K, listener: (this: XMLHttpRequestUpload, ev: XMLHttpRequestEventTargetEventMap[K]) => any, options?: boolean | AddEventListenerOptions): void;
    private handleReadyStateChange;
    private execute;
    /**
     * send fires off the XHR and returns a promise with the Result.
     * Multiple calls will result in the same return value (wrapped in Promise).
     */
    send(): Promise<Result<Response<T>, ClientError<E>>>;
    /**
     * Construct a client manually for a request with a response shaped T,
     * or in the case of a bad request, shaped E.
     *
     * `Client.create` is useful when you need a cancellation token
     * or need to listen to special XHR events (like upload progress).
     */
    static create<T = any, E = any>(method: string, url: string, options?: RequestOptions): Client<T, E>;
    /**
     * Make a GET request with a response shaped T,
     * or in the case of a bad request, shaped E.
     */
    static get<T = any, E = any>(url: string, query?: QStr.QueryObject): Promise<Result<Response<T>, ClientError<E>>>;
    /**
     * Make a POST request with a response shaped T,
     * or in the case of a bad request, shaped E.
     */
    static post<T = any, E = any>(url: string, data: RequestData, options?: RequestOptions): Promise<Result<Response<T>, ClientError<E>>>;
    /**
     * Make a PUT request with a response shaped T,
     * or in the case of a bad request, shaped E.
     */
    static put<T = any, E = any>(url: string, data: RequestData, options?: RequestOptions): Promise<Result<Response<T>, ClientError<E>>>;
    /**
     * Make a PATCH request with a response shaped T,
     * or in the case of a bad request, shaped E.
     */
    static patch<T = any, E = any>(url: string, data: RequestData, options?: RequestOptions): Promise<Result<Response<T>, ClientError<E>>>;
    /**
     * Make a DELETE request with a response shaped T,
     * or in the case of a bad request, shaped E.
     */
    static delete<T = any, E = any>(url: string, query?: QStr.QueryObject): Promise<Result<Response<T>, ClientError<E>>>;
    /**
     * Sets the type of the response to enable correct parsing of the
     * XHR response.
     */
    static setResponseType(type: XMLHttpRequestResponseType): void;
}
